/*
 * %CopyrightBegin%
 *
 * SPDX-License-Identifier: Apache-2.0
 *
 * Copyright Ericsson AB 1998-2025. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * %CopyrightEnd%
 */
#include <string.h>
#include "eidef.h"
#include "eiext.h"
#include "putget.h"

int ei_encode_boolean(char *buf, int *index, int p)
{
  char *s = buf + *index;
  char *s0 = s;
  char *val;
  int len;

  val = p ? "true" : "false";
  len = strlen(val);

  if (!buf) s += 2;
  else {
      put8(s, ERL_SMALL_ATOM_UTF8_EXT);
      put8(s, len);

      memcpy(s,val,len); /* unterminated string */
  }
  s += len;

  *index += s-s0; 

  return 0; 
}

